using System.Runtime.CompilerServices;
using Frontend;
using Frontend.AssetStores;
using Frontend.BackgroundServices;
using Frontend.Configuration;
using Frontend.Marketplaces;
using Frontend.Marketplaces.Interfaces;
using Frontend.Metrics;
using Nethereum.Web3;
using Nethereum.Web3.Accounts;
using Microsoft.EntityFrameworkCore;
using DotNetEnv;

using OpenTelemetry.Metrics;
using OpenTelemetry.Resources;
using Org.BouncyCastle.Utilities;
using System.Numerics;

// ensure the env file is loaded
Env.Load("../.env");
// the location differ in dev and releas mode.
var tokenAbiLocation = Environment.GetEnvironmentVariable("TOKEN_ABI_LOCATION")!;
var tokenAbi = File.ReadAllText(tokenAbiLocation)!;

var chainApiUrl = Environment.GetEnvironmentVariable("API_URL")!;
var accountPrivateKey = Environment.GetEnvironmentVariable("ACCOUNT_PRIVATE_KEY")!;
var marketplaceAccountPrivateKey = Environment.GetEnvironmentVariable("MARKETPLACE_ACCOUNT_PRIVATE_KEY")!;
var postgresHost = Environment.GetEnvironmentVariable("POSTGRES_HOST")!;
var postgresUser = Environment.GetEnvironmentVariable("POSTGRES_USER")!;
var postgresPassword = Environment.GetEnvironmentVariable("POSTGRES_PASSWORD")!;
var postgresDb = Environment.GetEnvironmentVariable("POSTGRES_DB")!;
var prometheusBaseUrl =Environment.GetEnvironmentVariable("PROMETHEUS_BASE_URL")!;

var builder = WebApplication.CreateBuilder(args);

// Configurations
var chainConfiguration = builder.Configuration.GetSection("Chain");
var chainSettings = new ChainSettings();
chainConfiguration.Bind(chainSettings);

// Add services to the container.
builder.Services.AddHostedService<PrometheusService>();
builder.Services.AddHostedService<RebalanceService>();
builder.Services.AddSingleton<Erc20TokenMetrics>();
builder.Services.AddSingleton<PotatoStorageMetrics>();
builder.Services.AddSingleton<PotatoMarketplaceMetrics>();
builder.Services.AddSingleton<RebalancerMetrics>();

Account account = new Account(
    accountPrivateKey,
    chainId: chainSettings.ChainId
);

Account marketplaceAccount = new(
    marketplaceAccountPrivateKey,
    chainId: chainSettings.ChainId
);

builder.Services.AddSingleton(provider => chainSettings);


builder.Services.AddSingleton(provider => account);
builder.Services.AddTransient<Web3>(sp =>
{
    var chainSettings = sp.GetRequiredService<ChainSettings>();
    var account = sp.GetRequiredService<Account>();
    return new Web3(account, chainApiUrl);
});

builder.Services.AddTransient<PrometheusClient>(sp => new PrometheusClient(prometheusBaseUrl));

builder.Services.AddSingleton<PoolMetrics>();
builder.Services.AddSingleton<IPotatoMarketPlace>(sp =>
{
    var chainSettings = sp.GetRequiredService<ChainSettings>();
    var web3Client = new Web3(marketplaceAccount, chainApiUrl);
    return new PotatoMarketplaceMock(marketplaceAccount, web3Client, chainSettings.WethTokenAddress);
});

builder.Services.AddDbContextFactory<PotatoStorage>(options =>
    options.UseNpgsql($"Host={postgresHost};Database={postgresDb};Username={postgresUser};Password={postgresPassword}"));

builder.Services.AddRazorPages();
builder.Services.AddOpenTelemetry()
    .WithMetrics(builder =>
    {
        builder.AddPrometheusExporter();

        builder.AddMeter("Microsoft.AspNetCore.Hosting",
                         "Microsoft.AspNetCore.Server.Kestrel",
                         "RoestiCoin");
    });

builder.Services.AddSingleton<PotatoStorage>(sp =>
{
    var factory = sp.GetRequiredService<IDbContextFactory<PotatoStorage>>();
    return factory.CreateDbContext();
});
builder.Services.AddSingleton<Rebalancer>();

var app = builder.Build();

var potatoStorageFactory = app.Services.GetRequiredService<IDbContextFactory<PotatoStorage>>();
var potatoStorage = potatoStorageFactory.CreateDbContext();
potatoStorage.Database.EnsureCreated();

// Configure the HTTP request pipeline.
if (!app.Environment.IsDevelopment())
{
    app.UseExceptionHandler("/Error");
    // The default HSTS value is 30 days. You may want to change this for production scenarios, see https://aka.ms/aspnetcore-hsts.
    app.UseHsts();

    app.UseHttpsRedirection();
}

app.MapPrometheusScrapingEndpoint();
app.UseStaticFiles();

app.UseRouting();

app.UseAuthorization();

app.MapRazorPages();

app.Run();
