using System.Numerics;
using Frontend.Functions;
using Nethereum.Contracts;
using Nethereum.Contracts.Standards.ERC20.ContractDefinition;
using Nethereum.Web3;
using Nethereum.Web3.Accounts;

namespace Frontend;

public class TokenClient
{
    const int Decimals = 18;

    private readonly Web3 _web3;
    private readonly string _tokenAddress;

    public string TokenAddress => _tokenAddress;

    public TokenClient(
        Web3 web3,
        string tokenAddress
    )
    {
        _tokenAddress = tokenAddress;
        _web3 = web3;
    }

    public async Task<decimal> TotalSupply()
    {
        var totalSupplyHandler = _web3.Eth.GetContractQueryHandler<TotalSupplyFunction>();
        var totalSupplyFunction = new TotalSupplyFunction();
        var totalSupply = await totalSupplyHandler.QueryAsync<BigInteger>(_tokenAddress, totalSupplyFunction);
        return ConvertSmallUnitToBigUnit(totalSupply);
    }

    public async Task Mint(string recipientAddress, BigInteger amount, CancellationToken cancellationToken)
    {
        var mintHandler = _web3.Eth.GetContractTransactionHandler<MintFunction>();
        var mintFunction = new MintFunction()
        {
            To = recipientAddress,
            Amount = amount,
        };

        await mintHandler.SendRequestAndWaitForReceiptAsync(_tokenAddress, mintFunction, cancellationToken);
    }

    public async Task Burn(BigInteger amount, CancellationToken cancellationToken)
    {
        var burnHandler = _web3.Eth.GetContractTransactionHandler<BurnFunction>();
        var burnFunction = new BurnFunction()
        {
            Amount = amount,
        };

        await burnHandler.SendRequestAndWaitForReceiptAsync(_tokenAddress, burnFunction, cancellationToken);
    }

    public decimal ConvertSmallUnitToBigUnit(BigInteger smallUnit)
    {
        return (decimal)(smallUnit / (BigInteger)Math.Pow(10, Decimals));
    }


    public async Task<BigInteger> GetAllowanceAsync(string ownerAddress, string spenderAddress)
    {
        var allowanceFunction = new AllowanceFunction
        {
            Owner = ownerAddress,
            Spender = spenderAddress
        };

        var handler = _web3.Eth.GetContractQueryHandler<AllowanceFunction>();
        var allowance = await handler.QueryAsync<BigInteger>(_tokenAddress, allowanceFunction);

        return allowance;
    }

    public async Task<string> ApproveMaxAsync(string spenderAddress)
    {
        return await ApproveAsync(spenderAddress, BigInteger.Pow(2, 256) - 1);
    }

    public async Task<string> ApproveAsync(string spenderAddress, BigInteger amount)
    {
        var approveFunction = new Nethereum.Contracts.Standards.ERC20.ContractDefinition.ApproveFunction
        {
            Spender = spenderAddress,
            Value = amount
        };

        var handler = _web3.Eth.GetContractTransactionHandler<Nethereum.Contracts.Standards.ERC20.ContractDefinition.ApproveFunction>();
        var transactionReceipt = await handler.SendRequestAndWaitForReceiptAsync(_tokenAddress, approveFunction);

        return transactionReceipt.TransactionHash;
    }

    public async Task<string> TransferAsync(string to, BigInteger amount)
    {
        var transferFunction = new TransferFunction
        {
            To = to,
            Value = amount
        };

        var handler = _web3.Eth.GetContractTransactionHandler<TransferFunction>();
        var transactionReceipt = await handler.SendRequestAndWaitForReceiptAsync(_tokenAddress, transferFunction);

        return transactionReceipt.TransactionHash;
    }

    public async Task<BigInteger> BalanceOf(string address){
        var func = new BalanceOfFunction
        {
            Owner = address,
        };

        var handler = _web3.Eth.GetContractQueryHandler<BalanceOfFunction>();
        return  await handler.QueryAsync<BigInteger>(_tokenAddress, func);

    }

}
